/*
 * Grundkurs Informatik - Beispielprogramm zum Hashing
 * H. Ernst, J. Schmidt 2016
 *
 * Hashing mit Überlaufbehandlung
 * durch verketteten Listen
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <stdlib.h>
#include <assert.h>

// Definition von bool für C90
typedef enum {false=0, true=1} bool;


// Anzahl Zeichen für zu speichernden String (inkl. Nullbyte)
#define DIM 20

// Größe des Arrays für Hashing
#define ANZ 100


void auflisten(void);

void einfuegen(void);

int hash(char *text);

void suchen(void);

void loeschen(void);

// Datensatz
struct rec {
    char info[DIM]; // Informationsteil
    struct rec *pnt; // Zeiger
};

struct rec *hfeld[ANZ]; // Hash-Feld, Zeiger


//**************************************************
// Hash-Funktion zur Bestimmung der Primäradresse
//--------------------------------------------------
int hash(char text[]) {
    int hashwert;
    int i;

    // hier eine ganz einfache Variante, gut zum Test
    // der Hashwert wird nur aus dem ersten Zeichen berechnet
    hashwert = text[0] % ANZ;

    // besser: Berechnung des Hashwerts aus allen Zeichen
    // for(i=0; i< strlen(text); i++) hashwert = (hashwert + text[i]) % ANZ;

    return hashwert;
}

//**************************************************
// Auflisten der Daten
//--------------------------------------------------
void auflisten(void) {
    int i;
    bool leer = true;
    struct rec *pnt;

    printf("\nDatensätze:\n");
    for (i = 0; i < ANZ; i++) {
        if (hfeld[i] != NULL) {
            leer = false;
            pnt = hfeld[i];
            while (pnt != NULL) {
                printf("%s ", pnt->info);
                pnt = pnt->pnt;
            }
        }
    }
    if (leer == true) {
        printf("Datei ist leer!\n");
    }

    printf("\n");
    return;
}

//**************************************************
// Suchen eines Datensatzes
//--------------------------------------------------
void suchen(void) {
    char text[DIM];
    struct rec *pnt;
    int hashwert;

    printf("\nGeben Sie den Datensatz ein: ");
    fgets(text, DIM, stdin);
    text[strlen(text) - 1] = '\0';   // \n aus String entfernen

    hashwert = hash(text);
    pnt = hfeld[hashwert]; // Primäradresse

    while (pnt != NULL && strcmp(text, pnt->info)) pnt = pnt->pnt;
    if (pnt != NULL) printf("\nGefunden: %s, Index: %d\n\n", pnt->info, hashwert);
    else printf("\nNicht gefunden: %s\n\n", text);

    return;
}

//**************************************************
// Löschen eines Datensatzes
//--------------------------------------------------
void loeschen(void) {
    char text[DIM];
    struct rec *pnt, *vor, *tmp;
    int hashwert;

    printf("\nGeben Sie den Datensatz ein: ");
    fgets(text, DIM, stdin);
    text[strlen(text) - 1] = '\0';   // \n aus String entfernen

    hashwert = hash(text);
    pnt = hfeld[hashwert]; // Primäradresse
    vor = NULL;  // Vorgänger in der Liste

    while (pnt != NULL && strcmp(text, pnt->info) != 0) {
        vor = pnt;
        pnt = pnt->pnt;
    }
    if (pnt != NULL) {
        printf("\nGelöscht: %s, Index: %d\n\n", pnt->info, hashwert);
        // Liste neu verketten...
        tmp = pnt;  // zu löschendes Element

        if(vor == NULL) {  // Sonderbehandlung für Listenanfang
            hfeld[hashwert] = tmp->pnt;
        }
        else {
            vor->pnt = tmp->pnt;
        }

        free(tmp);
    }
    else printf("\nNicht gefunden: %s\n\n", text);

    return;
}

//**************************************************
// Einfügen eines Datensatzes
//--------------------------------------------------
void einfuegen(void) {
    char text[DIM];
    int i;
    struct rec *neu, *pnt, *vor;
    bool ende = false;

    printf("\nEingabe, beenden mit <ENTER>");
    while(ende == false) {
        printf("\n> "); // Prompt
        fgets(text, DIM, stdin);

        if (text[0] == '\n') ende = true; // Eingabe beenden
        else {
            text[strlen(text) - 1] = '\0';   // \n aus String entfernen

            i = hash(text); // Primäradresse
            neu = (struct rec *) malloc(sizeof(struct rec));
            assert(neu != NULL);

            strcpy(neu->info, text);
            neu->pnt = NULL;

            // füge in lexikographischer Ordnung in verkettete Liste ein
            vor = NULL;
            pnt = hfeld[i];
            while (pnt != NULL && strcmp(text, pnt->info) > 0) {
                vor = pnt; // Einfügestelle
                pnt = pnt->pnt;
            }

            if(vor == NULL) {   // Sonderbehandlung für ersten Eintrag
                neu->pnt = hfeld[i];
                hfeld[i] = neu;
            }
            else {  // an passender Stelle in Liste einfügen
                vor->pnt = neu;
                neu->pnt = pnt;
            }
        }
    }
}


int main() {
    int i;
    char eingabe[DIM];
    bool ende = false;

    // Hash-Array initialisieren
    for(i = 0; i < ANZ; i++) hfeld[i] = NULL;

    printf("Hashing - Beenden mit <ENTER>\n");
    while(ende == false) {
        printf("Daten einfügen  - e\n");
        printf("Daten löschen   - l\n");
        printf("Daten suchen    - s\n");
        printf("Daten auflisten - a\n");
        fgets(eingabe, DIM, stdin);

        switch(eingabe[0]) {
            case 'e':
                einfuegen();
                break;
            case 'l':
                loeschen();
                break;
            case 's':
                suchen();
                break;
            case 'a':
                auflisten();
                break;
            case '\n':
                printf("Ende\n");
                ende = true;
                break;
            default:
                printf("Eingabe ungültig\n");
                break;
        }
    }

    return EXIT_SUCCESS;
}